module ManageIQ::Providers::Azure::ManagerMixin
  extend ActiveSupport::Concern

  def connect(options = {})
    raise MiqException::MiqHostError, _("No credentials defined") if missing_credentials?(options[:auth_type])

    client_id  = options[:user] || authentication_userid(options[:auth_type])
    client_key = options[:pass] || authentication_password(options[:auth_type])

    self.class.raw_connect(client_id, client_key, azure_tenant_id, subscription, options[:proxy_uri] || http_proxy_uri, provider_region, default_endpoint)
  end

  def verify_credentials(_auth_type = nil, options = {})
    connect(options)
  end

  def edit_with_params(params)
    default_endpoint = params.delete("endpoints").dig("default")
    default_authentication = params.delete("authentications").dig("default")

    tap do |ems|
      ems.default_authentication.assign_attributes(default_authentication)
      ems.default_endpoint.assign_attributes(default_endpoint)

      ems.assign_attributes(params)

      ems.save!
    end
  end

  module ClassMethods
    def params_for_create
      @params_for_create ||= {
        :fields => [
          {
            :component  => "select-field",
            :name       => "provider_region",
            :label      => _("Region"),
            :isRequired => true,
            :validate   => [{:type => "required-validator"}],
            :options    => ManageIQ::Providers::Azure::Regions.all.sort_by { |r| r[:description] }.map do |region|
              {
                :label => region[:description],
                :value => region[:name]
              }
            end
          },
          {
            :component  => "text-field",
            :name       => "uid_ems",
            :label      => _("Tenant ID"),
            :isRequired => true,
            :validate   => [{:type => "required-validator"}],
          },
          {
            :component  => "text-field",
            :name       => "subscription",
            :label      => _("Subscription ID"),
            :isRequired => true,
            :validate   => [{:type => "required-validator"}],
          },
          {
            :component => 'sub-form',
            :name      => 'endpoints',
            :title     => _("Endpoint"),
            :fields    => [
              {
                :component              => 'validate-provider-credentials',
                :name                   => 'authentications.default.valid',
                :validationDependencies => %w[type zone_id provider_region subscription uid_ems],
                :fields                 => [
                  {
                    :component => "text-field",
                    :name      => "endpoints.default.url",
                    :label     => _("Endpoint URL"),
                  },
                  {
                    :component  => "text-field",
                    :name       => "authentications.default.userid",
                    :label      => _("Client ID"),
                    :helperText => _("Should have privileged access, such as root or administrator."),
                    :isRequired => true,
                    :validate   => [{:type => "required-validator"}]
                  },
                  {
                    :component  => "password-field",
                    :name       => "authentications.default.password",
                    :label      => _("Client Key"),
                    :type       => "password",
                    :isRequired => true,
                    :validate   => [{:type => "required-validator"}]
                  },
                ],
              },
            ],
          },
        ],
      }.freeze
    end

    def create_from_params(params)
      endpoints = params.delete("endpoints") || {'default' => {}} # Fall back to an empty default endpoint
      authentications = params.delete("authentications")

      new(params).tap do |ems|
        endpoints.each do |authtype, endpoint|
          ems.endpoints.new(endpoint.merge(:role => authtype))
        end

        authentications.each do |authtype, authentication|
          ems.authentications.new(authentication.merge(:authtype => authtype))
        end

        ems.save!
      end
    end

    # Verify Credentials
    # args:
    # {
    #   "uid_ems"      => "",
    #   "subscription" => "",
    #   "region"       => "",
    #   "endpoints"    => {
    #     "default" => {
    #       "userid"   => "",
    #       "password" => "",
    #       "url"      => ""
    #     }
    #   }
    # }
    def verify_credentials(args)
      region           = args["region"]
      subscription     = args["subscription"]
      azure_tenant_id  = args["uid_ems"]
      default_endpoint = args.dig("authentications", "default")
      endpoint_url = args.dig("endpoints", "default", "url")

      client_id, client_key = default_endpoint&.values_at("userid", "password")

      client_key = MiqPassword.try_decrypt(client_key)
      # Pull out the password from the database if a provider ID is available
      client_key ||= find(args["id"]).authentication_password('default')

      !!raw_connect(client_id, client_key, azure_tenant_id, subscription, http_proxy_uri, region, endpoint_url)
    end

    def raw_connect(client_id, client_key, azure_tenant_id, subscription, proxy_uri = nil, provider_region = nil, endpoint = nil)
      require 'azure-armrest'

      if subscription.blank?
        raise MiqException::MiqInvalidCredentialsError, _("Incorrect credentials - check your Azure Subscription ID")
      end

      if provider_region.blank?
        $azure_log.warn("No region selected. Validating credentials against public environment.")
      end

      endpoint_url = endpoint.respond_to?(:url) ? endpoint.url : endpoint.to_s

      if endpoint_url.present?
        begin
          environment = ::Azure::Armrest::Environment.discover(:url => endpoint_url, :proxy => proxy_uri)
        rescue SocketError
          raise MiqException::MiqUnreachableError, _("Invalid endpoint")
        end
      else
        environment = environment_for(provider_region)
      end

      ::Azure::Armrest::Configuration.log = $azure_log

      connection_rescue_block do
        ::Azure::Armrest::Configuration.new(
          :client_id       => client_id,
          :client_key      => ManageIQ::Password.try_decrypt(client_key),
          :tenant_id       => azure_tenant_id,
          :subscription_id => subscription,
          :proxy           => proxy_uri,
          :environment     => environment
        )
      end
    end

    def connection_rescue_block
      require 'azure-armrest'
      yield
    rescue ArgumentError => err
      raise MiqException::MiqInvalidCredentialsError, _("Incorrect credentials - %{error_message}") % {:error_message => err.message}
    rescue ::Azure::Armrest::UnauthorizedException, ::Azure::Armrest::BadRequestException
      raise MiqException::MiqInvalidCredentialsError, _("Incorrect credentials - check your Azure Client ID and Client Key")
    rescue MiqException::MiqInvalidCredentialsError
      raise # Raise before falling into catch-all block below
    rescue StandardError => err
      _log.error("Error Class=#{err.class.name}, Message=#{err.message}, Backtrace=#{err.backtrace}")
      raise err, _("Unexpected response returned from system: %{error_message}") % {:error_message => err.message}
    end

    def environment_for(region)
      case region
      when /germany/i
        ::Azure::Armrest::Environment::Germany
      when /usgov/i
        ::Azure::Armrest::Environment::USGovernment
      else
        ::Azure::Armrest::Environment::Public
      end
    end

    # Discovery

    # Create EmsAzure instances for all regions with instances
    # or images for the given authentication. Created EmsAzure instances
    # will automatically have EmsRefreshes queued up.  If this is a greenfield
    # discovery, we will at least add an EmsAzure for eastus
    def discover(clientid, clientkey, azure_tenant_id, subscription)
      new_emses = []

      all_emses = includes(:authentications)
      all_ems_names = all_emses.index_by(&:name)

      known_emses = all_emses.select { |e| e.authentication_userid == clientid }
      known_ems_regions = known_emses.index_by(&:provider_region)

      config = raw_connect(clientid, clientkey, azure_tenant_id, subscription)

      azure_res = ::Azure::Armrest::ResourceService.new(config)
      azure_res.api_version = Settings.ems.ems_azure.api_versions.resource

      azure_res.list_locations.each do |region|
        next if known_ems_regions.include?(region.name)
        next if vms_in_region(azure_res, region.name).count.zero? # instances
        # TODO: Check if images are == 0 and if so then skip
        new_emses << create_discovered_region(region.name, clientid, clientkey, azure_tenant_id, subscription, all_ems_names)
      end

      # at least create the Azure-eastus region.
      if new_emses.blank? && known_emses.blank?
        new_emses << create_discovered_region("eastus", clientid, clientkey, azure_tenant_id, subscription, all_ems_names)
      end

      EmsRefresh.queue_refresh(new_emses) if new_emses.present?

      new_emses
    end

    def discover_queue(clientid, clientkey, azure_tenant_id, subscription)
      MiqQueue.put(
        :class_name  => name,
        :method_name => "discover_from_queue",
        :args        => [clientid, ManageIQ::Password.encrypt(clientkey), azure_tenant_id, subscription]
      )
    end

    def vms_in_region(azure_res, region)
      filter = "resourceType eq 'Microsoft.Compute/virtualMachines' and location eq '#{region}'"
      azure_res.list_all(:all => true, :filter => filter)
    end

    def discover_from_queue(clientid, clientkey, azure_tenant_id, subscription)
      discover(clientid, ManageIQ::Password.decrypt(clientkey), azure_tenant_id, subscription)
    end

    def create_discovered_region(region_name, clientid, clientkey, azure_tenant_id, subscription, all_ems_names)
      name = "Azure-#{region_name}"
      name = "Azure-#{region_name} #{clientid}" if all_ems_names.key?(name)

      while all_ems_names.key?(name)
        name_counter = name_counter.to_i + 1 if defined?(name_counter)
        name = "Azure-#{region_name} #{name_counter}"
      end

      new_ems = create!(
        :name            => name,
        :provider_region => region_name,
        :zone            => Zone.default_zone,
        :uid_ems         => azure_tenant_id,
        :subscription    => subscription
      )
      new_ems.update_authentication(
        :default => {
          :userid   => clientid,
          :password => clientkey
        }
      )
      new_ems
    end
  end
end
