module ManageIQ::Providers::Nuage::ManagerMixin
  extend ActiveSupport::Concern

  module ClassMethods
    def params_for_create
      @params_for_create ||= {
        :title  => "Configure Nuage",
        :fields => [
          {
            :component  => "text-field",
            :name       => "endpoints.default.server",
            :label      => "Server Hostname/IP Address",
            :isRequired => true,
            :validate   => [{:type => "required-validator"}]
          },
          {
            :component  => "text-field",
            :name       => "endpoints.default.username",
            :label      => "Username",
            :isRequired => true,
            :validate   => [{:type => "required-validator"}]
          },
          {
            :component  => "text-field",
            :name       => "endpoints.default.password",
            :label      => "Password",
            :type       => "password",
            :isRequired => true,
            :validate   => [{:type => "required-validator"}]
          },
          {
            :component  => "text-field",
            :name       => "endpoints.default.port",
            :label      => "Port",
            :type       => "number",
            :isRequired => true,
            :validate   => [{:type => "required-validator"}]
          },
          {
            :component    => "text-field",
            :name         => "endpoints.default.protocol",
            :label        => "Security Protocol",
            :initialValue => "ssl-no-validation", # TODO: This should be a dropdown
            :isRequired   => true,
            :validate     => [{:type => "required-validator"}]
          }
        ]
      }.freeze
    end

    # Verify Credentials
    #
    # args: {
    #   "endpoints" => {
    #     "default" => {
    #       "username" => String,
    #       "password" => String,
    #       "server" => String,
    #       "port" => Integer,
    #       "protocol" => String
    #     }
    #   }
    def verify_credentials(args)
      default_endpoint = args.dig("endpoints", "default")

      username, password, server, port, protocol = default_endpoint&.values_at(
        "username", "password", "server", "port", "protocol"
      )

      endpoint_opts = {
        :protocol => protocol,
        :hostname => server,
        :api_port => port
      }

      !!raw_connect(username, password, endpoint_opts)
    end

    def raw_connect(username, password, endpoint_opts)
      protocol    = endpoint_opts[:protocol].strip if endpoint_opts[:protocol]
      hostname    = endpoint_opts[:hostname].strip
      api_port    = endpoint_opts[:api_port]
      # In case API port is represented as a string, ensure it has no whitespaces.
      api_port.strip! if api_port.kind_of?(String)
      api_version = endpoint_opts[:api_version].to_s.strip

      # TODO(miha-plesko): Update UI to never pass in invalid version string like '? string:v2 ?'
      raise MiqException::MiqInvalidCredentialsError, 'Invalid API Version' unless api_version_valid?(api_version)

      url = auth_url(protocol, hostname, api_port, api_version)
      $nuage_log.debug("Connecting to Nuage VSD with url #{url}")

      connection_rescue_block do
        ManageIQ::Providers::Nuage::NetworkManager::VsdClient.new(url, username, password)
      end
    end

    def base_url(protocol, server, port)
      scheme = %w(ssl ssl-with-validation).include?(protocol) ? "https" : "http"
      URI::Generic.build(:scheme => scheme, :host => server, :port => port).to_s
    end

    def auth_url(protocol, server, port, version)
      URI(base_url(protocol, server, port)).tap { |url| url.path = "/nuage/api/#{version}" }.to_s
    end

    def translate_exception(err)
      require 'excon'
      case err
      when Excon::Errors::Unauthorized
        MiqException::MiqInvalidCredentialsError.new("Login failed due to a bad username or password.")
      when Excon::Errors::Timeout
        MiqException::MiqUnreachableError.new("Login attempt timed out")
      when Excon::Errors::SocketError
        MiqException::MiqHostError.new("Socket error: #{err.message}")
      when MiqException::MiqInvalidCredentialsError, MiqException::MiqHostError
        err
      when Net::HTTPBadResponse
        MiqException::MiqEVMLoginError.new("Login failed due to a bad security protocol, hostname or port.")
      else
        MiqException::MiqEVMLoginError.new("Unexpected response returned from system: #{err.message}")
      end
    end

    def connection_rescue_block
      yield
    rescue => err
      miq_exception = translate_exception(err)

      $nuage_log.error("Error Class=#{err.class.name}, Message=#{err.message}")
      raise miq_exception
    end

    def api_version_valid?(value)
      !!(value =~ /^v\d+[_.]\d+/) # e.g. 'v5_0' or 'v5.0'
    end
  end

  def connect(options = {})
    raise "no credentials defined" if self.missing_credentials?(options[:auth_type])

    username = options[:user] || authentication_userid(options[:auth_type])
    password = options[:pass] || authentication_password(options[:auth_type])
    protocol = options[:protocol] || security_protocol
    server   = options[:ip] || address
    port     = options[:port] || self.port
    version  = options[:version] || api_version

    endpoint_opts = {:protocol => protocol, :hostname => server, :api_port => port, :api_version => version}
    self.class.raw_connect(username, password, endpoint_opts)
  end

  def verify_credentials(auth_type = nil, options = {})
    auth_type ||= 'default'

    raise MiqException::MiqInvalidCredentialsError, "Unsupported auth type: #{auth_type}" unless supports_authentication?(auth_type)
    raise MiqException::MiqHostError, "No credentials defined" if missing_credentials?(auth_type)

    options[:auth_type] = auth_type
    case auth_type.to_s
    when 'default' then verify_api_credentials(options)
    when 'amqp'    then verify_amqp_credentials(options)
    end
  end

  def event_monitor_options
    @event_monitor_options ||= begin
      amqp_auth = authentications.detect { |a| a.authtype == 'amqp' }
      {
        :ems                       => self,
        :urls                      => amqp_urls,
        :sasl_allow_insecure_mechs => true, # Only plain (insecure) mechanism currently supported
        :user                      => amqp_auth.userid,
        :password                  => amqp_auth.password
      }
    end
  end

  private

  def amqp_urls
    amqp_endpoints = endpoints.select { |e| e.role == 'amqp' || e.role.start_with?('amqp_fallback') }
    amqp_endpoints.map do |e|
      "#{e.hostname}:#{e.port}"
    end
  end

  def verify_api_credentials(options = {})
    self.class.connection_rescue_block do
      with_provider_connection(options) {}
      true
    end
  end

  def verify_amqp_credentials(_options = {})
    ManageIQ::Providers::Nuage::NetworkManager::EventCatcher::Stream.test_amqp_connection(event_monitor_options)
  end

  def stop_event_monitor_queue_on_change
    if event_monitor_class && !new_record? && (authentications.detect(&:changed?) || endpoints.detect(&:changed?))
      $nuage_log.info("EMS: [#{name}], Credentials or endpoints have changed, stopping Event Monitor. It will be restarted by the WorkerMonitor.")
      stop_event_monitor_queue
    end
  end

  def stop_event_monitor_queue_on_credential_change
    stop_event_monitor_queue_on_change
  end
end
